from __future__ import annotations

from dataclasses import dataclass
from typing import Any, NamedTuple, TypeVar

from auto_editor.utils.log import Log
from auto_editor.utils.types import Align, align, anchor, color, natural, number, pos

# start - When the clip starts in the timeline
# dur - The duration of the clip in the timeline before speed is applied
# offset - When from the source to start playing the media at

T = TypeVar("T", bound=type)


class Attr(NamedTuple):
    names: tuple[str, ...]
    coerce: Any
    default: Any


def parse_dataclass(
    attrs_str: str,
    dataclass: T,
    builder: list[Attr],
    log: Log,
    _vars: dict[str, int] = {},
    coerce_default: bool = False,
) -> T:
    ARG_SEP = ","
    KEYWORD_SEP = "="

    # Positional Arguments
    #    --rectangle 0,end,10,20,20,30,#000, ...
    # Keyword Arguments
    #    --rectangle start=0,end=end,x1=10, ...

    def _values(name: str, val, _type, _vars: dict[str, int], log: Log):
        if val is None:
            return None
        if name in ("x", "width"):
            return pos((val, _vars["width"]))
        elif name in ("y", "height"):
            return pos((val, _vars["height"]))

        if _type is int:
            for key, item in _vars.items():
                if val == key:
                    return item

        try:
            _type(val)
        except TypeError as e:
            log.error(e)
        except Exception:
            log.error(f"{name}: variable '{val}' is not defined.")

        return _type(val)

    kwargs: dict[str, Any] = {}
    for attr in builder:
        key = attr.names[0]
        if coerce_default:
            kwargs[key] = _values(key, attr.default, attr.coerce, _vars, log)
        else:
            kwargs[key] = attr.default

    if attrs_str == "":
        for k, v in kwargs.items():
            if v is None:
                log.error(f"'{k}' must be specified.")
        return dataclass(**kwargs)

    d_name = dataclass.__name__
    allow_positional_args = True

    for i, arg in enumerate(attrs_str.split(ARG_SEP)):
        if i + 1 > len(builder):
            log.error(f"{d_name} has too many arguments, starting with '{arg}'.")

        if KEYWORD_SEP in arg:
            allow_positional_args = False

            parameters = arg.split(KEYWORD_SEP)
            if len(parameters) > 2:
                log.error(f"{d_name} invalid syntax: '{arg}'.")

            key, val = parameters
            found = False
            for attr in builder:
                if key in attr.names:
                    kwargs[attr.names[0]] = _values(
                        attr.names[0], val, attr.coerce, _vars, log
                    )
                    found = True
                    break

            if not found:
                from difflib import get_close_matches

                keys = set()
                for attr in builder:
                    for name in attr.names:
                        keys.add(name)

                more = ""
                if matches := get_close_matches(key, keys):
                    more = f"\n    Did you mean:\n        {', '.join(matches)}"

                log.error(f"{d_name} got an unexpected keyword '{key}'\n{more}")

        elif allow_positional_args:
            key = builder[i].names[0]
            kwargs[key] = _values(key, arg, builder[i].coerce, _vars, log)
        else:
            log.error(f"{d_name} positional argument follows keyword argument.")

    for k, v in kwargs.items():
        if v is None:
            log.error(f"'{k}' must be specified.")
    return dataclass(**kwargs)


@dataclass
class VideoObj:
    start: int
    dur: int
    offset: int
    speed: float
    src: int
    stream: int


@dataclass
class AudioObj:
    start: int
    dur: int
    offset: int
    speed: float
    src: int
    stream: int


@dataclass
class TextObj:
    start: int
    dur: int
    content: str
    x: int
    y: int
    font: str
    size: int
    fill: str
    align: Align
    stroke: int
    strokecolor: str


@dataclass
class ImageObj:
    start: int
    dur: int
    src: str
    x: int
    y: int
    opacity: float
    anchor: str
    rotate: float


@dataclass
class RectangleObj:
    start: int
    dur: int
    x: int
    y: int
    width: int
    height: int
    anchor: str
    fill: str
    stroke: int
    strokecolor: str


@dataclass
class EllipseObj:
    start: int
    dur: int
    x: int
    y: int
    width: int
    height: int
    anchor: str
    fill: str
    stroke: int
    strokecolor: str


video_builder = [
    Attr(("start",), natural, None),
    Attr(("dur",), natural, None),
    Attr(("offset",), natural, None),
    Attr(("speed",), number, None),
    Attr(("stream", "track"), natural, 0),
]
audio_builder = video_builder


def content(val: str) -> str:
    return val.replace("\\n", "\n").replace("\\;", ",")


text_builder = [
    Attr(("start",), natural, None),
    Attr(("dur",), natural, None),
    Attr(("content",), content, None),
    Attr(("x",), int, "50%"),
    Attr(("y",), int, "50%"),
    Attr(("font",), str, "Arial"),
    Attr(("size",), natural, 55),
    Attr(("fill", "color"), str, "#FFF"),
    Attr(("align",), align, "left"),
    Attr(("stroke",), natural, 0),
    Attr(("strokecolor",), color, "#000"),
]

img_builder = [
    Attr(("start",), natural, None),
    Attr(("dur",), natural, None),
    Attr(("src",), str, None),
    Attr(("x",), int, "50%"),
    Attr(("y",), int, "50%"),
    Attr(("opacity",), number, 1),
    Attr(("anchor",), anchor, "ce"),
    Attr(("rotate",), number, 0),
]

rect_builder = [
    Attr(("start",), natural, None),
    Attr(("dur",), natural, None),
    Attr(("x",), int, None),
    Attr(("y",), int, None),
    Attr(("width",), int, None),
    Attr(("height",), int, None),
    Attr(("anchor",), anchor, "ce"),
    Attr(("fill", "color"), color, "#c4c4c4"),
    Attr(("stroke",), natural, 0),
    Attr(("strokecolor",), color, "#000"),
]
ellipse_builder = rect_builder



from __future__ import annotations

from fractions import Fraction
from typing import List, NamedTuple, Type, Union

from auto_editor.ffwrapper import FileInfo
from auto_editor.method import get_has_loud
from auto_editor.objects import (
    AudioObj,
    EllipseObj,
    ImageObj,
    RectangleObj,
    TextObj,
    VideoObj,
)
from auto_editor.utils.bar import Bar
from auto_editor.utils.chunks import Chunks, chunkify, chunks_len, merge_chunks
from auto_editor.utils.func import apply_margin, cook, set_range
from auto_editor.utils.log import Log
from auto_editor.utils.types import Margin

Visual = Type[Union[TextObj, ImageObj, RectangleObj, EllipseObj]]
VLayer = List[Union[VideoObj, Visual]]
VSpace = List[VLayer]

ALayer = List[AudioObj]
ASpace = List[ALayer]


class Clip(NamedTuple):
    start: int
    dur: int
    offset: int
    speed: float
    src: int


def clipify(chunks: Chunks, src: int, start: float) -> list[Clip]:
    clips: list[Clip] = []
    # Add "+1" to match how chunks are rendered in 22w18a
    i = 0
    for chunk in chunks:
        if chunk[2] != 99999:
            if i == 0:
                dur = chunk[1] - chunk[0] + 1
                offset = chunk[0]
            else:
                dur = chunk[1] - chunk[0]
                offset = chunk[0] + 1

            if not (len(clips) > 0 and clips[-1].start == round(start)):
                clips.append(Clip(round(start), dur, offset, chunk[2], src))
            start += dur / chunk[2]
            i += 1

    return clips


def make_av(
    all_clips: list[list[Clip]], inputs: list[FileInfo]
) -> tuple[VSpace, ASpace]:
    vclips: VSpace = []

    max_a = 0
    for inp in inputs:
        max_a = max(max_a, len(inp.audios))

    aclips: ASpace = [[] for a in range(max_a)]

    for clips, inp in zip(all_clips, inputs):
        if len(inp.videos) > 0:
            for clip in clips:
                vclip_ = VideoObj(
                    clip.start, clip.dur, clip.offset, clip.speed, clip.src, 0
                )
                if len(vclips) == 0:
                    vclips = [[vclip_]]
                vclips[0].append(vclip_)
        if len(inp.audios) > 0:
            for clip in clips:
                for a, _ in enumerate(inp.audios):
                    aclips[a].append(
                        AudioObj(
                            clip.start, clip.dur, clip.offset, clip.speed, clip.src, a
                        )
                    )

    return vclips, aclips


def make_layers(
    inputs: list[FileInfo],
    tb: Fraction,
    method: str,
    margin: Margin,
    _min_cut: str | int,
    _min_clip: str | int,
    cut_out: list[list[str]],
    add_in: list[list[str]],
    mark_silent: list[list[str]],
    mark_loud: list[list[str]],
    set_speed_for_range: list[tuple[float, str, str]],
    silent_speed: float,
    loud_speed: float,
    bar: Bar,
    temp: str,
    log: Log,
) -> tuple[Chunks, VSpace, ASpace]:
    start = 0.0
    all_clips: list[list[Clip]] = []
    all_chunks: list[Chunks] = []

    def seconds_to_ticks(val: int | str, tb: Fraction) -> int:
        if isinstance(val, str):
            return int(float(val) * tb)
        return val

    start_margin, end_margin = margin
    start_margin = seconds_to_ticks(start_margin, tb)
    end_margin = seconds_to_ticks(end_margin, tb)
    min_clip = seconds_to_ticks(_min_clip, tb)
    min_cut = seconds_to_ticks(_min_cut, tb)

    for i in range(len(inputs)):
        has_loud = get_has_loud(method, i, inputs, tb, bar, temp, log)
        has_loud_length = len(has_loud)

        if len(mark_loud) > 0:
            has_loud = set_range(has_loud, mark_loud, tb, loud_speed, log)

        if len(mark_silent) > 0:
            has_loud = set_range(has_loud, mark_silent, tb, silent_speed, log)

        has_loud = cook(has_loud, min_clip, min_cut)
        has_loud = apply_margin(has_loud, has_loud_length, start_margin, end_margin)

        # Remove small clips/cuts created by applying other rules.
        has_loud = cook(has_loud, min_clip, min_cut)

        speed_list = has_loud.astype(float)
        del has_loud
        del has_loud_length

        # WARN: This breaks if speed is allowed to be 0
        speed_list[speed_list == 1] = loud_speed
        speed_list[speed_list == 0] = silent_speed

        if len(cut_out) > 0:
            speed_list = set_range(speed_list, cut_out, tb, 99999, log)

        if len(add_in) > 0:
            speed_list = set_range(speed_list, add_in, tb, loud_speed, log)

        for item in set_speed_for_range:
            speed_list = set_range(speed_list, [list(item[1:])], tb, item[0], log)

        _chunks = chunkify(speed_list)
        all_chunks.append(_chunks)
        all_clips.append(clipify(_chunks, i, start))
        start += round(chunks_len(_chunks))

    vclips, aclips = make_av(all_clips, inputs)
    return merge_chunks(all_chunks), vclips, aclips



from __future__ import annotations

from dataclasses import dataclass
from fractions import Fraction

from auto_editor.ffwrapper import FileInfo
from auto_editor.make_layers import ASpace, Visual, VSpace, make_layers
from auto_editor.objects import (
    EllipseObj,
    ImageObj,
    RectangleObj,
    TextObj,
    VideoObj,
    ellipse_builder,
    img_builder,
    parse_dataclass,
    rect_builder,
    text_builder,
)
from auto_editor.utils.bar import Bar
from auto_editor.utils.chunks import Chunks
from auto_editor.utils.log import Log
from auto_editor.utils.types import Args


@dataclass
class Timeline:
    inputs: list[FileInfo]
    timebase: Fraction
    samplerate: int
    res: tuple[int, int]
    background: str
    v: VSpace
    a: ASpace
    chunks: Chunks | None = None

    @property
    def inp(self) -> FileInfo:
        return self.inputs[0]

    @property
    def end(self) -> int:
        end = 0
        for vclips in self.v:
            if len(vclips) > 0:
                v = vclips[-1]
                if isinstance(v, VideoObj):
                    end = max(end, max(1, round(v.start + (v.dur / v.speed))))
                else:
                    end = max(end, v.start + v.dur)
        for aclips in self.a:
            if len(aclips) > 0:
                a = aclips[-1]
                end = max(end, max(1, round(a.start + (a.dur / a.speed))))

        return end

    def out_len(self) -> float:
        out_len: float = 0
        for vclips in self.v:
            dur: float = 0
            for v_obj in vclips:
                if isinstance(v_obj, VideoObj):
                    dur += v_obj.dur / v_obj.speed
                else:
                    dur += v_obj.dur
            out_len = max(out_len, dur)
        for aclips in self.a:
            dur = 0
            for aclip in aclips:
                dur += aclip.dur / aclip.speed
            out_len = max(out_len, dur)
        return out_len


def make_timeline(
    inputs: list[FileInfo], args: Args, sr: int, bar: Bar, temp: str, log: Log
) -> Timeline:

    if inputs:
        tb = inputs[0].get_fps() if args.frame_rate is None else args.frame_rate
        res = inputs[0].get_res() if args.resolution is None else args.resolution
    else:
        tb, res = Fraction(30), (1920, 1080)

    chunks, vclips, aclips = make_layers(
        inputs,
        tb,
        args.edit_based_on,
        args.frame_margin,
        args.min_cut_length,
        args.min_clip_length,
        args.cut_out,
        args.add_in,
        args.mark_as_silent,
        args.mark_as_loud,
        args.set_speed_for_range,
        args.silent_speed,
        args.video_speed,
        bar,
        temp,
        log,
    )

    timeline = Timeline(inputs, tb, sr, res, args.background, vclips, aclips, chunks)

    w, h = res
    _vars: dict[str, int] = {
        "width": w,
        "height": h,
        "start": 0,
        "end": timeline.end,
    }

    pool: list[Visual] = []
    for key, obj_str in args.pool:
        if key == "add_text":
            pool.append(
                parse_dataclass(obj_str, TextObj, text_builder, log, _vars, True)
            )
        if key == "add_rectangle":
            pool.append(
                parse_dataclass(obj_str, RectangleObj, rect_builder, log, _vars, True)
            )
        if key == "add_ellipse":
            pool.append(
                parse_dataclass(obj_str, EllipseObj, ellipse_builder, log, _vars, True)
            )
        if key == "add_image":
            pool.append(
                parse_dataclass(obj_str, ImageObj, img_builder, log, _vars, True)
            )

    for obj in pool:
        # Higher layers are visually on top
        # TODO: Use less layers.
        timeline.v.append([obj])

    return timeline



from __future__ import annotations

import os
import random
import re
from dataclasses import dataclass
from fractions import Fraction
from typing import Callable

import numpy as np
from numpy.typing import NDArray

from auto_editor.analyze.audio import audio_detection, audio_length
from auto_editor.analyze.motion import motion_detection
from auto_editor.analyze.pixeldiff import pixel_difference
from auto_editor.ffwrapper import FileInfo
from auto_editor.objects import Attr, parse_dataclass
from auto_editor.utils.bar import Bar
from auto_editor.utils.log import Log
from auto_editor.utils.types import Stream, natural, stream, threshold
from auto_editor.wavfile import read

BoolList = NDArray[np.bool_]
BoolOperand = Callable[[BoolList, BoolList], BoolList]


@dataclass
class Audio:
    threshold: float
    stream: Stream


@dataclass
class Motion:
    threshold: float
    stream: int
    blur: int
    width: int


@dataclass
class Pixeldiff:
    threshold: int
    stream: int


@dataclass
class Random:
    cutchance: float = 0.5
    seed: int = -1


audio_builder = [
    Attr(("threshold",), threshold, 0.04),
    Attr(("stream", "track"), stream, 0),
]
motion_builder = [
    Attr(("threshold",), threshold, 0.02),
    Attr(("stream", "track"), natural, 0),
    Attr(("blur",), natural, 9),
    Attr(("width",), natural, 400),
]
pixeldiff_builder = [
    Attr(("threshold",), natural, 1),
    Attr(("stream", "track"), natural, 0),
]
random_builder = [Attr(("cutchance",), threshold, 0.5), Attr(("seed",), int, -1)]


def get_media_length(path: str, i: int, tb: Fraction, temp: str, log: Log) -> int:
    # Read first audio track.

    audio_path = os.path.join(temp, f"{i}-0.wav")
    if os.path.isfile(audio_path):
        sr, samples = read(audio_path)
        return audio_length(len(samples), sr, tb, log)

    # If there's no audio, get length in video metadata.
    import av

    av.logging.set_level(av.logging.PANIC)
    cn = av.open(path, "r")
    if len(cn.streams.video) < 1:
        log.error("Could not get media duration")

    video = cn.streams.video[0]
    dur = int(video.duration * video.time_base * tb)
    log.debug(f"Video duration: {dur}")
    return dur


def get_audio_list(
    i: int,
    stream: int,
    threshold: float,
    tb: Fraction,
    bar: Bar,
    temp: str,
    log: Log,
) -> BoolList:
    if os.path.isfile(path := os.path.join(temp, f"{i}-{stream}.wav")):
        sr, samples = read(path)
    else:
        raise TypeError(f"Audio stream '{stream}' does not exist.")

    audio_list = audio_detection(samples, sr, tb, bar, log)
    return np.fromiter((x > threshold for x in audio_list), dtype=np.bool_)


def operand_combine(a: BoolList, b: BoolList, call: BoolOperand) -> BoolList:
    if len(a) > len(b):
        b = np.resize(b, len(a))
    if len(b) > len(a):
        a = np.resize(a, len(b))

    return call(a, b)


def get_all_list(path: str, i: int, tb: Fraction, temp: str, log: Log) -> BoolList:
    return np.zeros(get_media_length(path, i, tb, temp, log) - 1, dtype=np.bool_)


def get_stream_data(
    method: str,
    attrs_str: str,
    i: int,
    inputs: list[FileInfo],
    timebase: Fraction,
    bar: Bar,
    temp: str,
    log: Log,
) -> BoolList:

    inp = inputs[0]
    strict = len(inputs) < 2

    if method == "none":
        return np.ones(
            get_media_length(inp.path, i, timebase, temp, log) - 1, dtype=np.bool_
        )
    if method == "all":
        return get_all_list(inp.path, i, timebase, temp, log)
    if method == "random":
        robj = parse_dataclass(attrs_str, Random, random_builder, log)
        if robj.seed == -1:
            robj.seed = random.randint(0, 2147483647)

        l = get_media_length(inp.path, i, timebase, temp, log) - 1
        random.seed(robj.seed)
        log.debug(f"Seed: {robj.seed}")

        a = random.choices((0, 1), weights=(robj.cutchance, 1 - robj.cutchance), k=l)

        return np.asarray(a, dtype=np.bool_)
    if method == "audio":
        audio = parse_dataclass(attrs_str, Audio, audio_builder, log)
        if audio.stream == "all":
            total_list: NDArray[np.bool_] | None = None
            for s in range(len(inp.audios)):
                try:
                    audio_list = get_audio_list(
                        i, s, audio.threshold, timebase, bar, temp, log
                    )
                    if total_list is None:
                        total_list = audio_list
                    else:
                        total_list = operand_combine(
                            total_list, audio_list, np.logical_or
                        )
                except TypeError as e:
                    if not strict:
                        return get_all_list(inp.path, i, timebase, temp, log)
                    log.error(e)

            if total_list is None:
                if not strict:
                    return get_all_list(inp.path, i, timebase, temp, log)
                log.error("Input has no audio streams.")
            return total_list
        else:
            try:
                return get_audio_list(
                    i, audio.stream, audio.threshold, timebase, bar, temp, log
                )
            except TypeError as e:
                if not strict:
                    return get_all_list(inp.path, i, timebase, temp, log)
                log.error(e)
    if method == "motion":
        mobj = parse_dataclass(attrs_str, Motion, motion_builder, log)

        if mobj.stream >= len(inp.videos):
            if not strict:
                return get_all_list(inp.path, i, timebase, temp, log)
            log.error(f"Video stream '{mobj.stream}' does not exist.")

        motion_list = motion_detection(
            inp.path, mobj.stream, timebase, bar, mobj.width, mobj.blur
        )
        return np.fromiter((x >= mobj.threshold for x in motion_list), dtype=np.bool_)

    if method == "pixeldiff":
        pobj = parse_dataclass(attrs_str, Pixeldiff, pixeldiff_builder, log)

        if pobj.stream >= len(inp.videos):
            if not strict:
                return get_all_list(inp.path, i, timebase, temp, log)
            log.error(f"Video stream '{pobj.stream}' does not exist.")

        pixel_list = pixel_difference(inp.path, pobj.stream, timebase, bar)
        return np.fromiter((x >= pobj.threshold for x in pixel_list), dtype=np.bool_)

    raise ValueError(f"Unreachable. {method=}")


def get_has_loud(
    token_str: str,
    i: int,
    inputs: list[FileInfo],
    timebase: Fraction,
    bar: Bar,
    temp: str,
    log: Log,
) -> NDArray[np.bool_]:

    METHOD_ATTRS_SEP = ":"
    METHODS = ("audio", "motion", "pixeldiff", "random", "none", "all")

    result_array: NDArray[np.bool_] | None = None
    operand: str | None = None

    logic_funcs: dict[str, BoolOperand] = {
        "and": np.logical_and,
        "or": np.logical_or,
        "xor": np.logical_xor,
    }

    # See: https://stackoverflow.com/questions/1059559/
    for token in filter(None, re.split(r"[ _]+", token_str)):
        if METHOD_ATTRS_SEP in token:
            token, attrs_str = token.split(METHOD_ATTRS_SEP)
            if token not in METHODS:
                log.error(f"'{token}': Token not allowed to have attributes.")
        else:
            attrs_str = ""

        if token in METHODS:
            if result_array is not None and operand is None:
                log.error("Logic operator must be between two editing methods.")

            stream_data = get_stream_data(
                token, attrs_str, i, inputs, timebase, bar, temp, log
            )

            if operand == "not":
                result_array = np.logical_not(stream_data)
                operand = None
            elif result_array is None:
                result_array = stream_data
            elif operand is not None and operand in ("and", "or", "xor"):
                result_array = operand_combine(
                    result_array, stream_data, logic_funcs[operand]
                )
                operand = None

        elif token in ("and", "or", "xor"):
            if operand is not None:
                log.error("Invalid Editing Syntax.")
            if result_array is None:
                log.error(f"'{token}' operand needs two arguments.")
            operand = token
        elif token == "not":
            if operand is not None:
                log.error("Invalid Editing Syntax.")
            operand = token
        else:
            log.error(f"Unknown method/operator: '{token}'")

    if operand is not None:
        log.error(f"Dangling operand: '{operand}'")

    assert result_array is not None
    return result_array



from __future__ import annotations

from fractions import Fraction
from typing import overload

import numpy as np
from numpy.typing import NDArray

from auto_editor.utils.log import Log
from auto_editor.utils.types import time

"""
To prevent duplicate code being pasted between scripts, common functions should be
put here. Every function should be pure with no side effects.
"""


def to_timecode(secs: float | Fraction, fmt: str) -> str:
    sign = ""
    if secs < 0:
        sign = "-"
        secs = -secs

    _m, _s = divmod(secs, 60)
    _h, _m = divmod(_m, 60)
    s, m, h = float(_s), int(_m), int(_h)

    if fmt == "webvtt":
        if h == 0:
            return f"{sign}{m:02d}:{s:06.3f}"
        return f"{sign}{h:02d}:{m:02d}:{s:06.3f}"
    if fmt == "mov_text":
        return f"{sign}{h:02d}:{m:02d}:" + f"{s:06.3f}".replace(".", ",", 1)
    if fmt == "standard":
        return f"{sign}{h:02d}:{m:02d}:{s:06.3f}"
    if fmt == "ass":
        return f"{sign}{h:d}:{m:02d}:{s:05.2f}"
    if fmt == "rass":
        return f"{sign}{h:d}:{m:02d}:{s:02.0f}"

    raise ValueError("to_timecode: Unreachable")


def remove_small(
    has_loud: NDArray[np.bool_], lim: int, replace: int, with_: int
) -> NDArray[np.bool_]:
    start_p = 0
    active = False
    for j, item in enumerate(has_loud):
        if item == replace:
            if not active:
                start_p = j
                active = True
            # Special case for end.
            if j == len(has_loud) - 1:
                if j - start_p < lim:
                    has_loud[start_p : j + 1] = with_
        else:
            if active:
                if j - start_p < lim:
                    has_loud[start_p:j] = with_
                active = False
    return has_loud


@overload
def set_range(
    arr: NDArray[np.float_],
    range_syntax: list[list[str]],
    tb: Fraction,
    with_: float,
    log: Log,
) -> NDArray[np.float_]:
    pass


@overload
def set_range(
    arr: NDArray[np.bool_],
    range_syntax: list[list[str]],
    tb: Fraction,
    with_: float,
    log: Log,
) -> NDArray[np.bool_]:
    pass


def set_range(arr, range_syntax, tb, with_, log):
    def replace_variables_to_values(val: str, tb: Fraction, log: Log) -> int:
        if val == "start":
            return 0
        if val == "end":
            return len(arr)

        try:
            value = time(val)
        except TypeError as e:
            log.error(e)
        if isinstance(value, int):
            return value
        return round(float(value) * tb)

    for _range in range_syntax:
        pair = []
        for val in _range:
            num = replace_variables_to_values(val, tb, log)
            if num < 0:
                num += len(arr)
            pair.append(num)
        arr[pair[0] : pair[1]] = with_
    return arr


def cook(has_loud: NDArray[np.bool_], min_clip: int, min_cut: int) -> NDArray[np.bool_]:
    has_loud = remove_small(has_loud, min_clip, replace=1, with_=0)
    has_loud = remove_small(has_loud, min_cut, replace=0, with_=1)
    return has_loud


def apply_margin(
    has_loud: NDArray[np.bool_], has_loud_length: int, start_m: int, end_m: int
) -> NDArray[np.bool_]:

    # Find start and end indexes.
    start_index = []
    end_index = []
    for j in range(1, has_loud_length):
        if has_loud[j] != has_loud[j - 1]:
            if has_loud[j]:
                start_index.append(j)
            else:
                end_index.append(j)

    # Apply margin
    if start_m > 0:
        for i in start_index:
            has_loud[max(i - start_m, 0) : i] = True
    if start_m < 0:
        for i in start_index:
            has_loud[i : min(i - start_m, has_loud_length)] = False

    if end_m > 0:
        for i in end_index:
            has_loud[i : min(i + end_m, has_loud_length)] = True
    if end_m < 0:
        for i in end_index:
            has_loud[max(i + end_m, 0) : i] = False

    return has_loud


def merge(start_list: np.ndarray, end_list: np.ndarray) -> NDArray[np.bool_]:
    result = np.zeros((len(start_list)), dtype=np.bool_)

    for i, item in enumerate(start_list):
        if item == True:
            where = np.where(end_list[i:])[0]
            if len(where) > 0:
                result[i : where[0]] = True
    return result


def get_stdout(cmd: list[str]) -> str:
    from subprocess import PIPE, STDOUT, Popen

    stdout, _ = Popen(cmd, stdout=PIPE, stderr=STDOUT).communicate()
    return stdout.decode("utf-8", "replace")


def aspect_ratio(width: int, height: int) -> tuple[int, int]:
    if height == 0:
        return (0, 0)

    def gcd(a: int, b: int) -> int:
        while b:
            a, b = b, a % b
        return a

    c = gcd(width, height)
    return width // c, height // c


def human_readable_time(time_in_secs: float) -> str:
    units = "seconds"
    if time_in_secs >= 3600:
        time_in_secs = round(time_in_secs / 3600, 1)
        if time_in_secs % 1 == 0:
            time_in_secs = round(time_in_secs)
        units = "hours"
    if time_in_secs >= 60:
        time_in_secs = round(time_in_secs / 60, 1)
        if time_in_secs >= 10 or time_in_secs % 1 == 0:
            time_in_secs = round(time_in_secs)
        units = "minutes"
    return f"{time_in_secs} {units}"


def open_with_system_default(path: str, log: Log) -> None:
    import sys
    from subprocess import run

    if sys.platform == "win64" or sys.platform == "win32":
        from os import startfile

        try:
            startfile(path)
        except OSError:
            log.warning("Could not find application to open file.")
    else:
        try:  # should work on MacOS and some Linux distros
            run(["open", path])
        except Exception:
            try:  # should work on WSL2
                run(["cmd.exe", "/C", "start", path])
            except Exception:
                try:  # should work on most other Linux distros
                    run(["xdg-open", path])
                except Exception:
                    log.warning("Could not open output file.")


def append_filename(path: str, val: str) -> str:
    from os.path import splitext

    root, ext = splitext(path)
    return root + val + ext



from __future__ import annotations

import os.path
from dataclasses import dataclass
from math import ceil
from subprocess import DEVNULL, PIPE
from typing import Dict, Tuple, Union

import av
from PIL import Image, ImageChops, ImageDraw, ImageFont, ImageOps

from auto_editor.ffwrapper import FFmpeg
from auto_editor.make_layers import Visual
from auto_editor.objects import EllipseObj, ImageObj, RectangleObj, TextObj, VideoObj
from auto_editor.output import video_quality
from auto_editor.timeline import Timeline
from auto_editor.utils.bar import Bar
from auto_editor.utils.container import Container
from auto_editor.utils.encoder import encoders
from auto_editor.utils.log import Log
from auto_editor.utils.types import Args

av.logging.set_level(av.logging.PANIC)


@dataclass
class VideoFrame:
    index: int
    src: int


# From: github.com/PyAV-Org/PyAV/blob/main/av/video/frame.pyx
allowed_pix_fmt = {
    "yuv420p",
    "yuvj420p",
    "rgb48be",
    "rgb48le",
    "rgb64be",
    "rgb64le",
    "rgb24",
    "bgr24",
    "argb",
    "rgba",
    "abgr",
    "bgra",
    "gray",
    "gray8",
    "gray16be",
    "gray16le",
    "rgb8",
    "bgr8",
    "pal8",
}


def apply_anchor(
    x: int, y: int, width: int, height: int, anchor: str
) -> tuple[int, int]:
    if anchor == "ce":
        x = int((x * 2 - width) / 2)
        y = int((y * 2 - height) / 2)
    if anchor == "tr":
        x -= width
    if anchor == "bl":
        y -= height
    if anchor == "br":
        x -= width
        y -= height
    # Pillow uses 'tl' by default
    return x, y


def one_pos_two_pos(
    x: int, y: int, width: int, height: int, anchor: str
) -> tuple[int, int, int, int]:
    """Convert: x, y, width, height -> x1, y1, x2, y2"""

    if anchor == "ce":
        x1 = x - int(width / 2)
        x2 = x + int(width / 2)
        y1 = y - int(height / 2)
        y2 = y + int(height / 2)

        return x1, y1, x2, y2

    if anchor in ("tr", "br"):
        x1 = x - width
        x2 = x
    else:
        x1 = x
        x2 = x + width

    if anchor in ("tl", "tr"):
        y1 = y
        y2 = y + height
    else:
        y1 = y
        y2 = y - height

    return x1, y1, x2, y2


def render_av(
    ffmpeg: FFmpeg,
    timeline: Timeline,
    args: Args,
    bar: Bar,
    ctr: Container,
    temp: str,
    log: Log,
) -> tuple[str, bool]:

    FontCache = Dict[
        str, Tuple[Union[ImageFont.FreeTypeFont, ImageFont.ImageFont], float]
    ]

    font_cache: FontCache = {}
    img_cache: dict[str, Image.Image] = {}
    for layer in timeline.v:
        for vobj in layer:
            if isinstance(vobj, TextObj) and (vobj.font, vobj.size) not in font_cache:
                try:
                    if vobj.font == "default":
                        font_cache[(vobj.font, vobj.size)] = ImageFont.load_default()
                    else:
                        font_cache[(vobj.font, vobj.size)] = ImageFont.truetype(
                            vobj.font, vobj.size
                        )
                except OSError:
                    log.error(f"Font '{vobj.font}' not found.")

            if isinstance(vobj, ImageObj) and vobj.src not in img_cache:
                source = Image.open(vobj.src)
                source = source.convert("RGBA")
                source = source.rotate(vobj.rotate, expand=True)
                source = ImageChops.multiply(
                    source,
                    Image.new(
                        "RGBA", source.size, (255, 255, 255, int(vobj.opacity * 255))
                    ),
                )
                img_cache[vobj.src] = source

    inp = timeline.inp
    cns = [av.open(inp.path, "r") for inp in timeline.inputs]

    target_pix_fmt = "yuv420p"  # Reasonable default
    decoders = []
    tous = []
    seek_cost = []
    for c, cn in enumerate(cns):
        if len(cn.streams.video) == 0:
            decoders.append(None)
            tous.append(0)
            seek_cost.append(0)
        else:
            stream = cn.streams.video[0]
            stream.thread_type = "AUTO"

            if args.no_seek or stream.average_rate is None:
                sc_val = 4294967295  # 2 ** 32 - 1
                tou = 0
            else:
                # Keyframes are usually spread out every 5 seconds or less.
                sc_val = int(stream.average_rate * 5)
                tou = int(stream.time_base.denominator / stream.average_rate)

            seek_cost.append(sc_val)
            tous.append(tou)
            decoders.append(cn.decode(stream))

            if c == 0:
                target_pix_fmt = stream.pix_fmt

    log.debug(f"Tous: {tous}")
    log.debug(f"Clips: {timeline.v}")

    target_pix_fmt = target_pix_fmt if target_pix_fmt in allowed_pix_fmt else "yuv420p"
    log.debug(f"Target pix_fmt: {target_pix_fmt}")

    apply_video_later = True

    if args.video_codec in encoders:
        apply_video_later = encoders[args.video_codec]["pix_fmt"].isdisjoint(
            allowed_pix_fmt
        )

    if args.scale != 1:
        apply_video_later = False

    log.debug(f"apply video quality settings now: {not apply_video_later}")

    width, height = timeline.res
    spedup = os.path.join(temp, "spedup0.mp4")

    cmd = [
        "-hide_banner",
        "-y",
        "-f",
        "rawvideo",
        "-c:v",
        "rawvideo",
        "-pix_fmt",
        target_pix_fmt,
        "-s",
        f"{width}*{height}",
        "-framerate",
        f"{timeline.timebase}",
        "-i",
        "-",
        "-pix_fmt",
        target_pix_fmt,
    ]

    if apply_video_later:
        cmd.extend(["-c:v", "mpeg4", "-qscale:v", "1"])
    else:
        cmd = video_quality(cmd, args, inp, ctr)

    # Setting SAR requires re-encoding so we do it here.
    if timeline.inputs and timeline.inputs[0].videos:
        if (sar := timeline.inputs[0].videos[0].sar) is not None:
            cmd.extend(["-vf", f"setsar={sar.replace(':', '/')}"])

    cmd.append(spedup)

    process2 = ffmpeg.Popen(cmd, stdin=PIPE, stdout=DEVNULL, stderr=DEVNULL)
    assert process2.stdin is not None

    # First few frames can have an abnormal keyframe count, so never seek there.
    seek = 10

    seek_frame = None
    frames_saved = 0

    bar.start(timeline.end, "Creating new video")

    null_img = Image.new("RGB", (width, height), args.background)
    null_frame = av.VideoFrame.from_image(null_img).reformat(format=target_pix_fmt)

    frame_index = -1
    try:
        for index in range(timeline.end):
            # Add objects to obj_list
            obj_list: list[VideoFrame | Visual] = []
            for layer in timeline.v:
                for lobj in layer:
                    if isinstance(lobj, VideoObj):
                        if index >= lobj.start and index < lobj.start + ceil(
                            lobj.dur / lobj.speed
                        ):
                            obj_list.append(
                                VideoFrame(
                                    lobj.offset
                                    + round((index - lobj.start) * lobj.speed),
                                    lobj.src,
                                )
                            )
                    elif index >= lobj.start and index < lobj.start + lobj.dur:
                        obj_list.append(lobj)

            # Render obj_list
            frame = null_frame
            for obj in obj_list:
                if isinstance(obj, VideoFrame):
                    my_stream = cns[obj.src].streams.video[0]
                    if frame_index > obj.index:
                        log.debug(f"Seek: {frame_index} -> 0")
                        cns[obj.src].seek(0)
                        try:
                            frame = next(decoders[obj.src])
                            frame_index = round(frame.time * timeline.timebase)
                        except StopIteration:
                            pass

                    while frame_index < obj.index:
                        # Check if skipping ahead is worth it.
                        if (
                            obj.index - frame_index > seek_cost[obj.src]
                            and frame_index > seek
                        ):
                            seek = frame_index + (seek_cost[obj.src] // 2)
                            seek_frame = frame_index
                            log.debug(f"Seek: {frame_index} -> {obj.index}")
                            cns[obj.src].seek(
                                obj.index * tous[obj.src],
                                stream=my_stream,
                            )

                        try:
                            frame = next(decoders[obj.src])
                            frame_index = round(frame.time * timeline.timebase)
                        except StopIteration:
                            log.debug(f"No source frame at {index=}. Using null frame")
                            frame = null_frame
                            break

                        if seek_frame is not None:
                            log.debug(
                                f"Skipped {frame_index - seek_frame} frame indexes"
                            )
                            frames_saved += frame_index - seek_frame
                            seek_frame = None
                        if frame.key_frame:
                            log.debug(f"Keyframe {frame_index} {frame.pts}")

                    if frame.width != width or frame.height != height:
                        img = frame.to_image().convert("RGB")
                        if img.width > width or img.height > height:
                            factor = ceil(max(width / img.width, height / img.height))
                            img = ImageOps.scale(img, factor)
                        img = ImageOps.pad(img, (width, height), color=args.background)
                        frame = frame.from_image(img).reformat(format=target_pix_fmt)

                # Render visual objects
                else:
                    img = frame.to_image().convert("RGBA")
                    obj_img = Image.new("RGBA", img.size, (255, 255, 255, 0))
                    draw = ImageDraw.Draw(obj_img)

                    if isinstance(obj, TextObj):
                        text_w, text_h = draw.textsize(
                            obj.content, font=font_cache[(obj.font, obj.size)]
                        )
                        pos = apply_anchor(obj.x, obj.y, text_w, text_h, "ce")
                        draw.text(
                            pos,
                            obj.content,
                            font=font_cache[(obj.font, obj.size)],
                            fill=obj.fill,
                            align=obj.align,
                            stroke_width=obj.stroke,
                            stroke_fill=obj.strokecolor,
                        )

                    if isinstance(obj, RectangleObj):
                        draw.rectangle(
                            one_pos_two_pos(
                                obj.x, obj.y, obj.width, obj.height, obj.anchor
                            ),
                            fill=obj.fill,
                            width=obj.stroke,
                            outline=obj.strokecolor,
                        )

                    if isinstance(obj, EllipseObj):
                        draw.ellipse(
                            one_pos_two_pos(
                                obj.x, obj.y, obj.width, obj.height, obj.anchor
                            ),
                            fill=obj.fill,
                            width=obj.stroke,
                            outline=obj.strokecolor,
                        )

                    if isinstance(obj, ImageObj):
                        img_w, img_h = img_cache[obj.src].size
                        pos = apply_anchor(obj.x, obj.y, img_w, img_h, obj.anchor)
                        obj_img.paste(img_cache[obj.src], pos)

                    img = Image.alpha_composite(img, obj_img)
                    frame = frame.from_image(img).reformat(format=target_pix_fmt)

            if frame.format.name != target_pix_fmt:
                frame = frame.reformat(format=target_pix_fmt)
            process2.stdin.write(frame.to_ndarray().tobytes())

            if index % 3 == 0:
                bar.tick(index)

        bar.end()
        process2.stdin.close()
        process2.wait()
    except (OSError, BrokenPipeError):
        bar.end()
        ffmpeg.run_check_errors(cmd, log, True)
        log.error("FFmpeg Error!")

    log.debug(f"Total frames saved seeking: {frames_saved}")

    if args.scale != 1:
        sped_input = os.path.join(temp, "spedup0.mp4")
        spedup = os.path.join(temp, "scale0.mp4")
        scale_filter = f"scale=iw*{args.scale}:ih*{args.scale}"

        cmd = ["-i", sped_input, "-vf", scale_filter, spedup]

        check_errors = ffmpeg.pipe(cmd)
        if "Error" in check_errors or "failed" in check_errors:
            if "-allow_sw 1" in check_errors:
                cmd.insert(-1, "-allow_sw")
                cmd.insert(-1, "1")
            # Run again to show errors even if it might not work.
            ffmpeg.run(cmd)

    return spedup, apply_video_later



# type: ignore
from __future__ import annotations

import logging
import os
import platform
import shutil
import subprocess
import sys
from dataclasses import dataclass, field
from time import perf_counter
from typing import Callable, NoReturn
from fractions import Fraction

import av
import numpy as np

from auto_editor.vanparse import ArgumentParser

av.logging.set_level(av.logging.PANIC)


@dataclass
class TestArgs:
    only: list[str] = field(default_factory=list)
    help: bool = False
    category: str = "cli"


def test_options(parser):
    parser.add_argument("--only", "-n", nargs="*")
    parser.add_required(
        "category",
        nargs=1,
        choices=["cli", "sub", "api", "unit", "all"],
        help="Set what category of tests to run.",
    )
    return parser


def pipe_to_console(cmd: list[str]) -> tuple[int, str, str]:
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = process.communicate()
    return process.returncode, stdout.decode("utf-8"), stderr.decode("utf-8")


def cleanup(the_dir: str) -> None:
    for item in os.listdir(the_dir):
        item = os.path.join(the_dir, item)
        if (
            "_ALTERED" in item
            or item.endswith(".xml")
            or item.endswith(".fcpxml")
            or item.endswith(".mlt")
        ):
            os.remove(item)
        if item.endswith("_tracks"):
            shutil.rmtree(item)


def clean_all() -> None:
    cleanup("resources")
    cleanup(os.getcwd())


def get_runner() -> list[str]:
    if platform.system() == "Windows":
        return ["py", "-m", "auto_editor"]
    return ["python3", "-m", "auto_editor"]


def run_program(cmd: list[str]) -> None:
    no_open = "." in cmd[0]
    cmd = get_runner() + cmd

    if no_open:
        cmd += ["--no_open"]

    returncode, stdout, stderr = pipe_to_console(cmd)
    if returncode > 0:
        raise Exception(f"{stdout}\n{stderr}\n")


def check_for_error(cmd: list[str], match=None) -> None:
    returncode, stdout, stderr = pipe_to_console(get_runner() + cmd)
    if returncode > 0:
        if "Error!" in stderr:
            if match is not None and match not in stderr:
                raise Exception(f'Could\'t find "{match}"')
        else:
            raise Exception(f"Program crashed.\n{stdout}\n{stderr}")
    else:
        raise Exception("Program should not respond with a code 0.")


def make_np_list(in_file: str, compare_file: str, the_speed: float) -> None:
    from auto_editor.render.tsm.phasevocoder import phasevocoder
    from auto_editor.wavfile import read

    _, sped_chunk = read(in_file)
    channels = 2

    spedup_audio = phasevocoder(channels, the_speed, sped_chunk)
    loaded = np.load(compare_file)

    if not np.array_equal(spedup_audio, loaded["a"]):
        if spedup_audio.shape == loaded["a"].shape:
            print(f"Both shapes ({spedup_audio.shape}) are same")
        else:
            print(spedup_audio.shape)
            print(loaded["a"].shape)

        result = np.subtract(spedup_audio, loaded["a"])

        print(f"result non-zero: {np.count_nonzero(result)}")
        print(f"len of spedup_audio: {len(spedup_audio)}")

        print(
            np.count_nonzero(result) / spedup_audio.shape[0],
            "difference between arrays",
        )

        raise Exception(f"file {compare_file} doesn't match array.")

    # np.savez_compressed(out_file, a=spedup_audio)


class Tester:
    def __init__(self, args: TestArgs) -> None:
        self.passed_tests = 0
        self.failed_tests = 0
        self.args = args

    def run(self, func: Callable, cleanup=None, allow_fail=False) -> None:
        if self.args.only != [] and func.__name__ not in self.args.only:
            return

        start = perf_counter()
        try:
            func()
            end = perf_counter() - start
        except KeyboardInterrupt:
            print(f"Testing Interrupted by User.")
            clean_all()
            sys.exit(1)
        except Exception as e:
            self.failed_tests += 1
            print(f"Test '{func.__name__}' failed.\n{e}")
            if not allow_fail:
                logging.error("", exc_info=True)
                clean_all()
                sys.exit(1)
        else:
            self.passed_tests += 1
            print(f"Test '{func.__name__}' passed: {round(end, 2)} secs")
            if cleanup is not None:
                cleanup()

    def end(self) -> NoReturn:
        print(f"{self.passed_tests}/{self.passed_tests + self.failed_tests}")
        clean_all()
        sys.exit(0)


def main(sys_args: list[str] | None = None):
    if sys_args is None:
        sys_args = sys.argv[1:]

    args = test_options(ArgumentParser("test")).parse_args(TestArgs, sys_args)

    ### Tests ###

    ## API Tests ##

    def read_api_0_1():
        check_for_error(
            ["resources/json/0.1-non-zero-start.json"],
            "Error! First chunk must start with 0",
        )
        check_for_error(
            ["resources/json/0.1-disjoint.json"], "Error! Chunk disjointed at"
        )

    def help_tests():
        """check the help option, its short, and help on options and groups."""
        run_program(["--help"])
        run_program(["-h"])
        run_program(["--frame_margin", "--help"])
        run_program(["--frame_margin", "-h"])
        run_program(["--help", "--help"])
        run_program(["-h", "--help"])
        run_program(["--help", "-h"])
        run_program(["-h", "--help"])

    def version_test():
        """Test version flags and debug by itself."""
        run_program(["--version"])
        run_program(["-v"])
        run_program(["-V"])
        run_program(["--debug"])

    def parser_test():
        check_for_error(["example.mp4", "--video-speed"], "needs argument")

    def tsm_1a5_test():
        make_np_list(
            "resources/wav/example-cut-s16le.wav",
            "resources/data/example_1.5_speed.npz",
            1.5,
        )

    def tsm_2a0_test():
        make_np_list(
            "resources/wav/example-cut-s16le.wav",
            "resources/data/example_2.0_speed.npz",
            2,
        )

    def info():
        run_program(["info", "example.mp4"])
        run_program(["info", "resources/only-video/man-on-green-screen.mp4"])
        run_program(["info", "resources/multi-track.mov"])
        run_program(["info", "resources/new-commentary.mp3"])
        run_program(["info", "resources/testsrc.mkv"])

    def levels():
        run_program(["levels", "resources/multi-track.mov"])
        run_program(["levels", "resources/new-commentary.mp3"])

    def subdump():
        run_program(["subdump", "resources/subtitle.mp4"])

    def grep():
        run_program(["grep", "boop", "resources/subtitle.mp4"])

    def desc():
        run_program(["desc", "example.mp4"])

    def example_tests():
        run_program(["example.mp4", "--video_codec", "uncompressed"])
        with av.open("example_ALTERED.mp4") as cn:
            video = cn.streams.video[0]
            assert video.average_rate == 30
            assert video.width == 1280
            assert video.height == 720
            assert video.codec.name == "mpeg4"
            assert cn.streams.audio[0].codec.name == "aac"
            assert cn.streams.audio[0].rate == 48000

        run_program(["example.mp4"])
        with av.open("example_ALTERED.mp4") as cn:
            video = cn.streams.video[0]
            assert video.average_rate == 30
            assert video.width == 1280
            assert video.height == 720
            assert video.codec.name == "h264"
            assert cn.streams.audio[0].codec.name == "aac"
            assert cn.streams.audio[0].rate == 48000
            assert video.language == "eng"
            assert cn.streams.audio[0].language == "eng"

    # PR #260
    def high_speed_test():
        run_program(["example.mp4", "--video-speed", "99998"])

    # Issue #184
    def unit_tests():
        run_program(
            ["example.mp4", "--mark_as_loud", "20s,22sec", "25secs,26.5seconds"]
        )
        run_program(["example.mp4", "--edit", "all", "--set-speed", "125%,-30,end"])
        run_program(["example.mp4", "--sample_rate", "44_100"])
        run_program(["example.mp4", "--margin", "3_0"])
        run_program(["example.mp4", "--sample_rate", "44100 Hz"])
        run_program(["example.mp4", "--sample_rate", "44.1 kHz"])
        run_program(["example.mp4", "--edit", "audio:threshold=4%"])

    def backwards_range_test():
        """
        Cut out the last 5 seconds of a media file by using negative number in the
        range.
        """
        run_program(["example.mp4", "--edit", "none", "--cut_out", "-5secs,end"])
        run_program(["example.mp4", "--edit", "all", "--add_in", "-5secs,end"])

    def cut_out_test():
        run_program(
            [
                "example.mp4",
                "--edit",
                "none",
                "--video_speed",
                "2",
                "--silent_speed",
                "3",
                "--cut_out",
                "2secs,10secs",
            ]
        )
        run_program(
            [
                "example.mp4",
                "--edit",
                "all",
                "--video_speed",
                "2",
                "--add_in",
                "2secs,10secs",
            ]
        )

    def gif_test():
        """
        Feed auto-editor a gif file and make sure it can spit out a correctly formated
        gif. No editing is requested.
        """
        run_program(["resources/only-video/man-on-green-screen.gif", "--edit", "none"])
        with av.open("resources/only-video/man-on-green-screen_ALTERED.gif") as cn:
            assert cn.streams.video[0].codec.name == "gif"

    def margin_tests():
        run_program(["example.mp4", "-m", "3"])
        run_program(["example.mp4", "--margin", "3"])
        run_program(["example.mp4", "-m", "0.3sec"])
        run_program(["example.mp4", "-m", "6f,-3secs"])
        run_program(["example.mp4", "-m", "3,5 frames"])
        run_program(["example.mp4", "-m", "0.4 seconds"])

    def input_extension():
        """Input file must have an extension. Throw error if none is given."""

        shutil.copy("example.mp4", "example")
        check_for_error(["example", "--no_open"], "must have an extension.")
        os.remove("example")

    def output_extension():
        # Add input extension to output name if no output extension is given.
        run_program(["example.mp4", "-o", "out"])
        with av.open("out.mp4") as cn:
            assert cn.streams.video[0].codec.name == "h264"

        os.remove("out.mp4")

        run_program(["resources/testsrc.mkv", "-o", "out"])
        with av.open("out.mkv") as cn:
            assert cn.streams.video[0].codec.name == "h264"

        os.remove("out.mkv")

    def progress_ops_test():
        run_program(["example.mp4", "--progress", "machine"])
        run_program(["example.mp4", "--progress", "none"])
        run_program(["example.mp4", "--progress", "ascii"])

    def silent_threshold():
        run_program(["resources/new-commentary.mp3", "--edit", "audio:threshold=0.1"])

    def track_tests():
        run_program(["resources/multi-track.mov", "--keep_tracks_seperate"])

    def json_tests():
        run_program(["example.mp4", "--export_as_json"])
        run_program(["example.json"])

    def resolution_and_scale():
        run_program(["example.mp4", "--scale", "1.5"])
        with av.open("example_ALTERED.mp4") as cn:
            assert cn.streams.video[0].average_rate == 30
            assert cn.streams.video[0].width == 1920
            assert cn.streams.video[0].height == 1080
            assert cn.streams.audio[0].rate == 48000

        run_program(["example.mp4", "--scale", "0.2"])
        with av.open("example_ALTERED.mp4") as cn:
            assert cn.streams.video[0].average_rate == 30
            assert cn.streams.video[0].width == 256
            assert cn.streams.video[0].height == 144
            assert cn.streams.audio[0].rate == 48000

        run_program(["example.mp4", "-res", "700,380", "-b", "darkgreen"])
        with av.open("example_ALTERED.mp4") as cn:
            assert cn.streams.video[0].average_rate == 30
            assert cn.streams.video[0].width == 700
            assert cn.streams.video[0].height == 380
            assert cn.streams.audio[0].rate == 48000

    def obj_makes_video():
        run_program(
            [
                "resources/new-commentary.mp3",
                "--add-rectangle",
                "0,30,0,0,300,300,fill=blue",
                "-o",
                "out.mp4",
            ]
        )
        with av.open("out.mp4") as cn:
            assert len(cn.streams.video) == 1
            assert len(cn.streams.audio) == 1
            assert cn.streams.video[0].width == 1920
            assert cn.streams.video[0].height == 1080
            assert cn.streams.video[0].average_rate == 30

    def various_errors_test():
        check_for_error(
            ["example.mp4", "--add_rectangle", "0,60", "--cut_out", "60,end"]
        )

    def render_video_objs():
        run_program(
            [
                "resources/testsrc.mp4",
                "--mark_as_loud",
                "start,end",
                "--add_rectangle",
                "0,30,0,200,100,300,fill=#43FA56,stroke=10",
            ]
        )
        os.remove("resources/testsrc_ALTERED.mp4")

        # Every element should be visible, order should be preserved.
        run_program(
            [
                "example.mp4",
                "--add-ellipse",
                "0,30,50%,50%,300,300,fill=red",
                "--add-rectangle",
                "0,30,500,440,400,200,fill=skyblue",
                "--add-ellipse",
                "0,30,50%,50%,100,100,fill=darkgreen",
                "--edit",
                "none",
                "--cut-out",
                "30,end",
            ]
        )

        # Both ellipses should be visible
        run_program(
            [
                "example.mp4",
                "--add-ellipse",
                "0,60,50%,50%,300,300,fill=darkgreen",
                "0,30,50%,50%,200,200,fill=green",
                "--edit",
                "none",
                "--cut-out",
                "60,end",
            ]
        )

    def render_text():
        run_program(["example.mp4", "--add-text", "0,30,This is my text,font=default"])

    def check_font_error():
        check_for_error(
            ["example.mp4", "--add-text", "0,30,text,0,0,notafont"], "not found"
        )

    def export_tests():
        for test_name in (
            "aac.m4a",
            "alac.m4a",
            "wav/pcm-f32le.wav",
            "wav/pcm-s32le.wav",
            "multi-track.mov",
            "subtitle.mp4",
            "testsrc.mkv",
        ):

            test_file = f"resources/{test_name}"
            run_program([test_file])
            run_program([test_file, "--edit", "none"])
            run_program([test_file, "-exp"])
            run_program([test_file, "-exf"])
            run_program([test_file, "-exs"])
            run_program([test_file, "--export_as_clip_sequence"])
            run_program([test_file, "--preview"])
            cleanup("resources")

    def codec_tests():
        run_program(["example.mp4", "--video_codec", "h264"])
        run_program(["example.mp4", "--audio_codec", "ac3"])

    def combine():
        run_program(["example.mp4", "--mark_as_silent", "0,171", "-o", "hmm.mp4"])
        run_program(["example.mp4", "hmm.mp4", "--combine-files", "--debug"])
        os.remove("hmm.mp4")

    # Issue #241
    def multi_track_edit():
        run_program(
            [
                "example.mp4",
                "resources/multi-track.mov",
                "--edit",
                "audio:stream=1",
                "-o",
                "out.mov",
            ]
        )
        with av.open("out.mov", "r") as cn:
            assert len(cn.streams.audio) == 1

    def concat_mux_tracks():
        run_program(["example.mp4", "resources/multi-track.mov", "-o", "out.mov"])
        with av.open("out.mov", "r") as cn:
            assert len(cn.streams.audio) == 1

    def concat_multiple_tracks():
        run_program(
            [
                "resources/multi-track.mov",
                "resources/multi-track.mov",
                "--keep-tracks-separate",
                "-o",
                "out.mov",
            ]
        )
        with av.open("out.mov", "r") as cn:
            assert len(cn.streams.audio) == 2, f"audio streams: {len(cn.streams.audio)}"

        run_program(
            [
                "example.mp4",
                "resources/multi-track.mov",
                "--keep-tracks-separate",
                "-o",
                "out.mov",
            ]
        )
        with av.open("out.mov", "r") as cn:
            assert len(cn.streams.audio) == 2
        os.remove("out.mov")

    def frame_rate():
        run_program(["example.mp4", "-r", "15", "--no-seek"])
        with av.open("example_ALTERED.mp4", "r") as cn:
            video = cn.streams.video[0]
            assert video.average_rate == 15
            dur = float(video.duration * video.time_base)
            assert dur - 17.33333333333333333333333 < 3

        run_program(["example.mp4", "-r", "20"])
        with av.open("example_ALTERED.mp4", "r") as cn:
            video = cn.streams.video[0]
            assert video.average_rate == 20
            dur = float(video.duration * video.time_base)
            assert dur - 17.33333333333333333333333 < 2

        run_program(["example.mp4", "-r", "60"])
        with av.open("example_ALTERED.mp4", "r") as cn:
            video = cn.streams.video[0]
            assert video.average_rate == 60
            dur = float(video.duration * video.time_base)
            assert dur - 17.33333333333333333333333 < 0.3

    def image_test():
        run_program(["resources/embedded-image/h264-png.mp4"])
        with av.open("resources/embedded-image/h264-png_ALTERED.mp4", "r") as cn:
            assert cn.streams.video[0].codec.name == "h264"
            assert cn.streams.video[1].codec.name == "png"

        run_program(["resources/embedded-image/h264-mjpeg.mp4"])
        with av.open("resources/embedded-image/h264-mjpeg_ALTERED.mp4", "r") as cn:
            assert cn.streams.video[0].codec.name == "h264"
            assert cn.streams.video[1].codec.name == "mjpeg"

        run_program(["resources/embedded-image/h264-png.mkv"])
        with av.open("resources/embedded-image/h264-png_ALTERED.mkv", "r") as cn:
            assert cn.streams.video[0].codec.name == "h264"
            assert cn.streams.video[1].codec.name == "png"

        run_program(["resources/embedded-image/h264-mjpeg.mkv"])
        with av.open("resources/embedded-image/h264-mjpeg_ALTERED.mkv", "r") as cn:
            assert cn.streams.video[0].codec.name == "h264"
            assert cn.streams.video[1].codec.name == "mjpeg"

    def motion_tests():
        run_program(
            [
                "resources/only-video/man-on-green-screen.mp4",
                "--edit",
                "motion",
                "--debug",
                "--frame_margin",
                "0",
                "-mcut",
                "0",
                "-mclip",
                "0",
            ]
        )
        run_program(
            [
                "resources/only-video/man-on-green-screen.mp4",
                "--edit",
                "motion:threshold=0",
            ]
        )

    def edit_positive_tests():
        run_program(["resources/multi-track.mov", "--edit", "audio:stream=all"])
        run_program(["resources/multi-track.mov", "--edit", "not audio:stream=all"])
        run_program(
            [
                "resources/multi-track.mov",
                "--edit",
                "not audio:threshold=4% or audio:stream=1",
            ]
        )
        # run_program(['resources/multi-track.mov', '--edit', 'not audio:threshold=4% or not audio:stream=1'])

    def edit_negative_tests():
        check_for_error(
            ["resources/wav/example-cut-s16le.wav", "--edit", "motion"],
            "Video stream '0' does not exist",
        )
        check_for_error(
            ["resources/only-video/man-on-green-screen.gif", "--edit", "audio"],
            "Audio stream '0' does not exist",
        )
        check_for_error(
            ["example.mp4", "--edit", "not"], "Error! Dangling operand: 'not'"
        )
        check_for_error(
            ["example.mp4", "--edit", "audio and"], "Error! Dangling operand: 'and'"
        )
        check_for_error(
            ["example.mp4", "--edit", "and"],
            "Error! 'and' operand needs two arguments.",
        )
        check_for_error(
            ["example.mp4", "--edit", "and audio"],
            "Error! 'and' operand needs two arguments.",
        )
        check_for_error(
            ["example.mp4", "--edit", "or audio"],
            "Error! 'or' operand needs two arguments.",
        )
        check_for_error(
            ["example.mp4", "--edit", "audio four audio"],
            "Error! Unknown method/operator: 'four'",
        )
        check_for_error(
            ["example.mp4", "--edit", "audio audio"],
            "Logic operator must be between two editing methods",
        )

    def yuv442p():
        run_program(["resources/test_yuv422p.mp4"])

    #  Issue 280
    def SAR():
        run_program(["resources/SAR-2by3.mp4"])
        with av.open("resources/SAR-2by3_ALTERED.mp4") as cn:
            assert cn.streams.video[0].sample_aspect_ratio == Fraction(2, 3)

    tests = []

    if args.category in ("unit", "all"):
        tests.extend([tsm_1a5_test, tsm_2a0_test])

    if args.category in ("api", "all"):
        tests.append(read_api_0_1)

    if args.category in ("sub", "all"):
        tests.extend([info, levels, subdump, grep, desc])

    if args.category in ("cli", "all"):
        tests.extend(
            [
                SAR,
                yuv442p,
                obj_makes_video,
                edit_positive_tests,
                edit_negative_tests,
                multi_track_edit,
                concat_mux_tracks,
                concat_multiple_tracks,
                render_video_objs,
                resolution_and_scale,
                various_errors_test,
                render_text,
                check_font_error,
                frame_rate,
                help_tests,
                version_test,
                parser_test,
                combine,
                example_tests,
                export_tests,
                high_speed_test,
                unit_tests,
                backwards_range_test,
                cut_out_test,
                image_test,
                gif_test,
                margin_tests,
                input_extension,
                output_extension,
                progress_ops_test,
                silent_threshold,
                track_tests,
                json_tests,
                codec_tests,
                motion_tests,
            ]
        )

    tester = Tester(args)

    for test in tests:
        tester.run(test)

    tester.end()


if __name__ == "__main__":
    main()



from __future__ import annotations

import os
import wave

from auto_editor.render.tsm.phasevocoder import phasevocoder
from auto_editor.timeline import Timeline
from auto_editor.utils.bar import Bar
from auto_editor.utils.log import Log
from auto_editor.wavfile import read


def make_new_audio(timeline: Timeline, bar: Bar, temp: str, log: Log) -> list[str]:
    samplerate = timeline.samplerate
    tb = timeline.timebase
    output = []
    samples = {}

    for l, layer in enumerate(timeline.a):
        bar.start(len(layer), "Creating new audio")

        # See: https://github.com/python/cpython/blob/3.10/Lib/wave.py
        path = os.path.join(temp, f"new{l}.wav")
        output.append(path)
        writer = wave.open(path, "wb")
        writer.setnchannels(2)
        writer.setframerate(samplerate)
        writer.setsampwidth(2)

        for c, clip in enumerate(layer):
            if f"{clip.src}-{clip.stream}" not in samples:
                audio_path = os.path.join(temp, f"{clip.src}-{clip.stream}.wav")
                assert os.path.exists(audio_path), f"{audio_path} Not found"
                samples[f"{clip.src}-{clip.stream}"] = read(audio_path)[1]

            samp_list = samples[f"{clip.src}-{clip.stream}"]

            samp_start = clip.offset * samplerate // tb
            samp_end = (clip.offset + clip.dur) * samplerate // tb
            if samp_end > len(samp_list):
                samp_end = len(samp_list)

            if clip.speed == 1:
                writer.writeframesraw(samp_list[samp_start:samp_end])  # type: ignore
            else:
                data = phasevocoder(2, clip.speed, samp_list[samp_start:samp_end])
                if data.shape[0] != 0:
                    writer.writeframesraw(data)  # type: ignore

            bar.tick(c)

        writer.close()
        bar.end()

    return output



from fractions import Fraction
from math import ceil

import numpy as np
import numpy.typing as npt

from auto_editor.utils.bar import Bar
from auto_editor.utils.log import Log


def get_max_volume(s: np.ndarray) -> float:
    return max(float(np.max(s)), -float(np.min(s)))


def audio_length(samp_count: int, sr: int, tb: Fraction, log: Log) -> int:
    samp_per_ticks = sr / tb
    ticks = ceil(samp_count / samp_per_ticks)
    log.debug(f"Audio Length: {ticks}")
    log.debug(f"... without ceil: {float(samp_count / samp_per_ticks)}")
    return ticks


def audio_detection(
    samples: np.ndarray, sr: int, tb: Fraction, bar: Bar, log: Log
) -> npt.NDArray[np.float_]:

    max_volume = get_max_volume(samples)
    log.debug(f"Max volume: {max_volume}")

    samp_count = samples.shape[0]
    samp_per_ticks = sr / tb

    audio_ticks = audio_length(samp_count, sr, tb, log)
    bar.start(audio_ticks, "Analyzing audio volume")

    threshold_list = np.zeros((audio_ticks), dtype=np.float_)

    if max_volume == 0:  # Prevent dividing by zero
        return threshold_list

    # Determine when audio is silent or loud.
    for i in range(audio_ticks):
        if i % 500 == 0:
            bar.tick(i)

        start = int(i * samp_per_ticks)
        end = min(int((i + 1) * samp_per_ticks), samp_count)

        threshold_list[i] = get_max_volume(samples[start:end]) / max_volume

    bar.end()
    return threshold_list



from __future__ import annotations

import os

from auto_editor.ffwrapper import FFmpeg, FileInfo
from auto_editor.timeline import Timeline, make_timeline
from auto_editor.utils.bar import Bar
from auto_editor.utils.chunks import Chunk, Chunks
from auto_editor.utils.container import Container, container_constructor
from auto_editor.utils.log import Log
from auto_editor.utils.types import Args


def set_output_name(path: str, inp_ext: str, export: str) -> str:
    root, ext = os.path.splitext(path)

    if export == "json":
        return f"{root}.json"
    if export == "final-cut-pro":
        return f"{root}.fcpxml"
    if export == "shotcut":
        return f"{root}.mlt"
    if export == "premiere":
        return f"{root}.xml"
    if export == "audio":
        return f"{root}_ALTERED.wav"
    if ext == "":
        return root + inp_ext

    return f"{root}_ALTERED{ext}"


codec_error = "'{}' codec is not supported in '{}' container."


def set_video_codec(
    codec: str, inp: FileInfo | None, out_ext: str, ctr: Container, log: Log
) -> str:
    if codec == "auto":
        codec = "h264" if (inp is None or not inp.videos) else inp.videos[0].codec
        if ctr.vcodecs is not None:
            if ctr.vstrict and codec not in ctr.vcodecs:
                return ctr.vcodecs[0]

            if codec in ctr.disallow_v:
                return ctr.vcodecs[0]
        return codec

    if codec == "copy":
        if inp is None:
            log.error("No input to copy its codec from.")
        if not inp.videos:
            log.error("Input file does not have a video stream to copy codec from.")
        codec = inp.videos[0].codec

    if codec == "uncompressed":
        codec = "mpeg4"

    if ctr.vstrict:
        assert ctr.vcodecs is not None
        if codec not in ctr.vcodecs:
            log.error(codec_error.format(codec, out_ext))

    if codec in ctr.disallow_v:
        log.error(codec_error.format(codec, out_ext))

    return codec


def set_audio_codec(
    codec: str, inp: FileInfo | None, out_ext: str, ctr: Container, log: Log
) -> str:
    if codec == "auto":
        codec = "aac" if (inp is None or not inp.audios) else inp.audios[0].codec
        if ctr.acodecs is not None:
            if ctr.astrict and codec not in ctr.acodecs:
                return ctr.acodecs[0]

            if codec in ctr.disallow_a:
                return ctr.acodecs[0]
        return codec

    if codec == "copy":
        if inp is None:
            log.error("No input to copy its codec from.")
        if not inp.audios:
            log.error("Input file does not have an audio stream to copy codec from.")
        codec = inp.audios[0].codec

    if codec != "unset":
        if ctr.astrict:
            assert ctr.acodecs is not None
            if codec not in ctr.acodecs:
                log.error(codec_error.format(codec, out_ext))

        if codec in ctr.disallow_a:
            log.error(codec_error.format(codec, out_ext))

    return codec


def edit_media(
    paths: list[str], ffmpeg: FFmpeg, args: Args, temp: str, log: Log
) -> str | None:

    bar = Bar(args.progress)
    timeline = None

    if paths:
        path_ext = os.path.splitext(paths[0])[1]
        if path_ext == ".json":
            from auto_editor.formats.json import read_json

            timeline = read_json(paths[0], ffmpeg, log)
            inputs: list[FileInfo] = timeline.inputs
        else:
            inputs = [FileInfo(path, ffmpeg, log) for path in paths]
    else:
        inputs = []
    del paths

    inp = None if not inputs else inputs[0]

    if inp is None:
        output = "out.mp4" if args.output_file is None else args.output_file
    else:
        if args.output_file is None:
            output = set_output_name(inp.path, inp.ext, args.export)
        else:
            output = args.output_file
            if os.path.splitext(output)[1] == "":
                output = set_output_name(output, inp.ext, args.export)

    out_ext = os.path.splitext(output)[1].replace(".", "")

    # Check if export options make sense.
    ctr = container_constructor(out_ext)

    if ctr.samplerate is not None and args.sample_rate not in ctr.samplerate:
        log.error(f"'{out_ext}' container only supports samplerates: {ctr.samplerate}")

    args.video_codec = set_video_codec(args.video_codec, inp, out_ext, ctr, log)
    args.audio_codec = set_audio_codec(args.audio_codec, inp, out_ext, ctr, log)

    if args.keep_tracks_separate and ctr.max_audios == 1:
        log.warning(f"'{out_ext}' container doesn't support multiple audio tracks.")

    if not args.preview and not args.timeline:
        if os.path.isdir(output):
            log.error("Output path already has an existing directory!")

        if os.path.isfile(output) and inputs[0].path != output:
            log.debug(f"Removing already existing file: {output}")
            os.remove(output)

    # Extract subtitles in their native format.
    if inp is not None and len(inp.subtitles) > 0:
        cmd = ["-i", inp.path, "-hide_banner"]
        for s, sub in enumerate(inp.subtitles):
            cmd.extend(["-map", f"0:s:{s}"])
        for s, sub in enumerate(inp.subtitles):
            cmd.extend([os.path.join(temp, f"{s}s.{sub.ext}")])
        ffmpeg.run(cmd)
        del inp

    log.conwrite("Extracting audio")

    cmd = []
    for i, inp in enumerate(inputs):
        cmd.extend(["-i", inp.path])
    cmd.append("-hide_banner")

    if args.sample_rate is None:
        if inputs:
            samplerate = inputs[0].get_samplerate()
        else:
            samplerate = 48000
    else:
        samplerate = args.sample_rate

    for i, inp in enumerate(inputs):
        for s in range(len(inp.audios)):
            cmd.extend(
                [
                    "-map",
                    f"{i}:a:{s}",
                    "-ac",
                    "2",
                    "-ar",
                    f"{samplerate}",
                    "-rf64",
                    "always",
                    os.path.join(temp, f"{i}-{s}.wav"),
                ]
            )

    ffmpeg.run(cmd)

    if timeline is None:
        timeline = make_timeline(inputs, args, samplerate, bar, temp, log)

    if args.timeline:
        from auto_editor.formats.json import make_json_timeline

        make_json_timeline(args.api, 0, timeline, log)
        return None

    if args.preview:
        from auto_editor.preview import preview

        preview(timeline, temp, log)
        return None

    if args.export == "json":
        from auto_editor.formats.json import make_json_timeline

        make_json_timeline(args.api, output, timeline, log)
        return output

    if args.export == "premiere":
        from auto_editor.formats.premiere import premiere_xml

        premiere_xml(temp, output, timeline)
        return output

    if args.export == "final-cut-pro":
        from auto_editor.formats.final_cut_pro import fcp_xml

        fcp_xml(output, timeline)
        return output

    if args.export == "shotcut":
        from auto_editor.formats.shotcut import shotcut_xml

        shotcut_xml(output, timeline)
        return output

    def make_media(timeline: Timeline, output: str) -> None:
        from auto_editor.output import mux_quality_media
        from auto_editor.render.video import render_av

        visual_output = []
        audio_output = []
        apply_later = False
        inp = timeline.inputs[0]

        if ctr.allow_subtitle:
            from auto_editor.render.subtitle import cut_subtitles

            cut_subtitles(ffmpeg, timeline, temp, log)

        if ctr.allow_audio:
            from auto_editor.render.audio import make_new_audio

            audio_output = make_new_audio(timeline, bar, temp, log)

        if ctr.allow_video:
            if len(timeline.v) > 0:
                out_path, apply_later = render_av(
                    ffmpeg, timeline, args, bar, ctr, temp, log
                )
                visual_output.append((True, out_path))

            for v, vid in enumerate(inp.videos, start=1):
                if ctr.allow_image and vid.codec in ("png", "mjpeg", "webp"):
                    out_path = os.path.join(temp, f"{v}.{vid.codec}")
                    # fmt: off
                    ffmpeg.run(["-i", inp.path, "-map", "0:v", "-map", "-0:V",
                        "-c", "copy", out_path])
                    # fmt: on
                    visual_output.append((False, out_path))

        log.conwrite("Writing output file")
        mux_quality_media(
            ffmpeg,
            visual_output,
            audio_output,
            apply_later,
            ctr,
            output,
            args,
            inp,
            temp,
            log,
        )

    if args.export == "clip-sequence":
        chunks = timeline.chunks
        if chunks is None:
            log.error("Timeline to complex to use clip-sequence export")

        from auto_editor.make_layers import clipify, make_av
        from auto_editor.utils.func import append_filename

        def pad_chunk(chunk: Chunk, total: int) -> Chunks:
            start = [] if chunk[0] == 0 else [(0, chunk[0], 99999.0)]
            end = [] if chunk[1] == total else [(chunk[1], total, 99999.0)]
            return start + [chunk] + end

        total_frames = chunks[-1][1] - 1
        clip_num = 0
        for chunk in chunks:
            if chunk[2] == 99999:
                continue

            _c = pad_chunk(chunk, total_frames)
            vspace, aspace = make_av([clipify(_c, 0, 0)], [inp])
            my_timeline = Timeline(
                timeline.inputs,
                timeline.timebase,
                timeline.samplerate,
                timeline.res,
                "#000",
                vspace,
                aspace,
                _c,
            )

            make_media(my_timeline, append_filename(output, f"-{clip_num}"))
            clip_num += 1
    else:
        make_media(timeline, output)
    return output



from __future__ import annotations

import difflib
import re
import sys
import textwrap
from dataclasses import dataclass
from shutil import get_terminal_size
from typing import Any, Callable, Iterator, Literal, Sequence, TypeVar, Union

import auto_editor
from auto_editor.utils.log import Log

T = TypeVar("T", bound=type)
Nargs = Union[int, Literal["*"]]


@dataclass
class Required:
    names: Sequence[str]
    nargs: Nargs = "*"
    type: type = str
    choices: Sequence[str] | None = None
    help: str = ""
    _type: str = "required"


@dataclass
class Options:
    names: Sequence[str]
    nargs: Nargs = 1
    type: type = str
    flag: bool = False
    choices: Sequence[str] | None = None
    pool: bool = False
    help: str = ""
    _type: str = "option"


@dataclass
class OptionText:
    text: str
    _type: str


def indent(text: str, prefix: str) -> str:
    def predicate(line: str) -> str:
        return line.strip()

    def prefixed_lines() -> Iterator[str]:
        for line in text.splitlines(True):
            yield (prefix + line if predicate(line) else line)

    return "".join(prefixed_lines())


def out(text: str) -> None:
    width = get_terminal_size().columns - 3

    indent_regex = re.compile(r"^(\s+)")
    wrapped_lines = []

    for line in text.split("\n"):
        exist_indent = re.search(indent_regex, line)
        pre_indent = exist_indent.groups()[0] if exist_indent else ""

        wrapped_lines.append(
            textwrap.fill(line, width=width, subsequent_indent=pre_indent)
        )

    print("\n".join(wrapped_lines))


def print_program_help(reqs: list[Required], args: list[Options | OptionText]) -> None:
    text = ""
    for arg in args:
        if isinstance(arg, OptionText):
            text += f"\n  {arg.text}\n" if arg._type == "text" else "\n"
        else:
            text += "  " + ", ".join(arg.names) + f": {arg.help}\n"
    text += "\n"
    for req in reqs:
        text += "  " + ", ".join(req.names) + f": {req.help}\n"
    out(text)


def get_help_data() -> dict[str, dict[str, str]]:
    import json
    import os.path

    dirpath = os.path.dirname(os.path.realpath(__file__))

    with open(os.path.join(dirpath, "help.json")) as fileobj:
        data = json.load(fileobj)

    assert isinstance(data, dict)
    return data


def to_underscore(name: str) -> str:
    """Convert new style options to old style.  e.g. --hello-world -> --hello_world"""
    return name[:2] + name[2:].replace("-", "_")


def to_key(op: Options | Required) -> str:
    """Convert option name to arg key.  e.g. --hello-world -> hello_world"""
    return op.names[0][:2].replace("-", "") + op.names[0][2:].replace("-", "_")


def print_option_help(program_name: str, ns_obj: T, option: Options) -> None:
    text = f"  {', '.join(option.names)}\n\n"

    bar_len = 11
    if option.flag:
        text += "    type: flag\n"
        bar_len = 15
    else:
        if option.nargs != 1:
            _add = f"    nargs: {option.nargs}\n"
            bar_len = len(_add)
            text += _add

        _add = f"    type: {option.type.__name__}\n"
        bar_len = len(_add)
        text += _add

        default: str | None = None
        try:
            default = getattr(ns_obj, to_key(option))
        except AttributeError:
            pass

        if default is not None:
            if isinstance(default, tuple):
                _add = f"    default: {','.join(map(str, default))}\n"
            else:
                _add = f"    default: {default}\n"
            bar_len = len(_add)
            text += _add

        if option.choices is not None:
            text += "    choices: " + ", ".join(option.choices) + "\n"

    text += f"    {'-' * (bar_len - 5)}\n\n    {option.help}\n\n"
    data = get_help_data()

    if option.names[0] in data[program_name]:
        text += indent(data[program_name][option.names[0]], "    ") + "\n"

    out(text)


def get_option(name: str, options: list[Options]) -> Options | None:
    for option in options:
        if name in option.names or name in map(to_underscore, option.names):
            return option
    return None


def parse_value(option: Options | Required, val: str | None) -> Any:
    if val is None and option.nargs == 1:
        Log().error(f"{option.names[0]} needs argument.")

    try:
        value = option.type(val)
    except TypeError as e:
        Log().error(e)

    if option.choices is not None and value not in option.choices:
        my_choices = ", ".join(option.choices)

        Log().error(
            f"{value} is not a choice for {option.names[0]}\nchoices are:\n  {my_choices}"
        )

    return value


class ArgumentParser:
    def __init__(self, program_name: str) -> None:
        self.program_name = program_name
        self.requireds: list[Required] = []
        self.options: list[Options] = []
        self.args: list[Options | OptionText] = []

    def add_argument(self, *args: str, **kwargs) -> None:
        x = Options(args, **kwargs)
        self.options.append(x)
        self.args.append(x)

    def add_required(self, *args: str, **kwargs) -> None:
        self.requireds.append(Required(args, **kwargs))

    def add_text(self, text: str) -> None:
        self.args.append(OptionText(text, "text"))

    def add_blank(self) -> None:
        self.args.append(OptionText("", "blank"))

    def parse_args(
        self,
        ns_obj: T,
        sys_args: list[str],
        macros: list[tuple[set[str], list[str]]] | None = None,
    ) -> T:
        if len(sys_args) == 0:
            out(get_help_data()[self.program_name]["_"])
            sys.exit()

        if len(sys_args) == 1 and sys_args[0] in ("-v", "-V"):
            sys.stdout.write(f"{auto_editor.version} ({auto_editor.__version__})\n")
            sys.exit()

        if macros is not None:
            _macros = [(x[0].union(map(to_underscore, x[0])), x[1]) for x in macros]
            for old_options, new in _macros:
                for old_option in old_options:
                    if old_option in sys_args:
                        pos = sys_args.index(old_option)
                        sys_args[pos : pos + 1] = new
            del _macros
        del macros

        ns = ns_obj()
        option_names: list[str] = []

        program_name = self.program_name
        requireds = self.requireds
        options = self.options
        args = self.args

        builtin_help = Options(
            ("--help", "-h"),
            flag=True,
            help="Show info about this program or option then exit.",
        )
        options.append(builtin_help)
        args.append(builtin_help)

        # Figure out command line options changed by user.
        used_options: list[Options] = []

        req_list: list[str] = []
        req_list_name = requireds[0].names[0]
        setting_req_list = requireds[0].nargs != 1

        option_list: Any = []
        op_key: str = ""
        oplist_name: str | None = None
        oplist_coerce: Callable[[str], str] = str

        i = 0
        while i < len(sys_args):
            arg = sys_args[i]
            option = get_option(arg, options)

            if option is None:
                if oplist_name == "pool":
                    option_list.append((op_key, arg))
                elif oplist_name is not None:
                    option_list.append(oplist_coerce(arg))
                elif requireds and not arg.startswith("--"):
                    if requireds[0].nargs == 1:
                        ns.__setattr__(req_list_name, parse_value(requireds[0], arg))
                        requireds.pop()
                    else:
                        req_list.append(arg)
                else:
                    label = "option" if arg.startswith("--") else "short"

                    # 'Did you mean' message might appear that options need a comma.
                    if arg.replace(",", "") in option_names:
                        Log().error(f"Option '{arg}' has an unnecessary comma.")

                    close_matches = difflib.get_close_matches(arg, option_names)
                    if close_matches:
                        Log().error(
                            f"Unknown {label}: {arg}\n\n    Did you mean:\n        "
                            + ", ".join(close_matches)
                        )
                    Log().error(f"Unknown {label}: {arg}")
            else:
                if oplist_name is not None and oplist_name != "pool":
                    ns.__setattr__(oplist_name, option_list)

                if not option.pool:
                    if option in used_options:
                        Log().error(f"Cannot repeat option {option.names[0]} twice.")
                    used_options.append(option)

                option_list = []
                oplist_name = None
                oplist_coerce = option.type

                key = to_key(option)

                next_arg = None if i == len(sys_args) - 1 else sys_args[i + 1]
                if next_arg in ("-h", "--help"):
                    print_option_help(program_name, ns_obj, option)
                    sys.exit()

                if option.nargs != 1:
                    if option.pool:
                        oplist_name = "pool"
                        op_key = key
                        ns.pool.append((key, next_arg))
                    else:
                        oplist_name = key
                        ns.__setattr__(oplist_name, parse_value(option, next_arg))
                elif option.flag:
                    ns.__setattr__(key, True)
                else:
                    ns.__setattr__(key, parse_value(option, next_arg))
                    i += 1
            i += 1

        if oplist_name == "pool":
            ns.pool += option_list
        elif oplist_name is not None:
            ns.__setattr__(oplist_name, option_list)

        if setting_req_list:
            ns.__setattr__(req_list_name, req_list)

        if ns.help:
            print_program_help(requireds, args)
            sys.exit()

        return ns
